/*
 * bltool - A Firmware Downloader/Uploader for ROSAND Technologies products
 * Copyright (C) 2000-2005  Brian S. Dean <bsd@bsdhome.com>
 * Copyright 2007-2009 Joerg Wunsch <j@uriah.heep.sax.de>
 * Copyright 2011, Rosand Technologies.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>

#include "bltool.h"
#include "config.h"
#include "pgm.h"
#include "8051.h"
#include "swra327.h"
#include "serial.h"

struct pdata
{
  unsigned char *flash_pagecache;
  unsigned int flash_pagesize;
};

#define NUM_TRY 3

#define PDATA(pgm) ((struct pdata *)(pgm->cookie))

static void swra327_setup(PROGRAMMER * pgm)
{
  if ((pgm->cookie = malloc(sizeof(struct pdata))) == 0) {
    fprintf(stderr,
	    "%s: swra327_setup(): Out of memory allocating private data\n",
	    progname);
    return;
  }
  memset(pgm->cookie, 0, sizeof(struct pdata));
}

static void swra327_teardown(PROGRAMMER * pgm)
{
  free(pgm->cookie);
}

static int swra327_send(PROGRAMMER * pgm, char * data, size_t len)
{
  unsigned char buf[256 + 4];              /* max MESSAGE_BODY of 66 bytes, 4 bytes overhead */
  unsigned char crc = 0;
  int i;

  memset(buf, 0xff, sizeof(buf));
  buf[0] = START_OF_FRAME;
  buf[1] = len-1;
  buf[2] = TOKEN;
  memcpy(buf+3, data, len);

  // calculate the XOR checksum
  for (i=0; i<3+len; i++)
	  crc ^= buf[i];
  buf[i++] = crc;

  return serial_send(&pgm->fd, buf, i);
}

static int swra327_recv(PROGRAMMER * pgm, char * buf, size_t len)
{
  return serial_recv(&pgm->fd, (unsigned char *)buf, len);
}

static int swra327_drain(PROGRAMMER * pgm, int display)
{
  return serial_drain(&pgm->fd, display);
}

static int swra327_vfy_cmd_sent(PROGRAMMER * pgm, unsigned char *data,  char * errmsg)
{
  unsigned char buf[275];
  unsigned char *tmp = buf;
  unsigned char c, crc = 0;
  int i;
  unsigned char len;

  if (swra327_recv(pgm, tmp, 2) < 0) {
    fprintf(stderr, "%s: %s: programmer did not respond to command.\n",
            progname, errmsg);
    return -1;
  }
  len = *(tmp+1);
  tmp +=2;

  swra327_recv(pgm, tmp, len+3);

  for (i=0, tmp=buf; i < len+4; i++) {
	  crc ^= *tmp++;
  }
  if (crc != *tmp) {
	  if (verbose)
		  fprintf(stderr, "%s: %s: reply error.\n", progname, errmsg);
	  return -1;
  }

  tmp = buf + 4;

  if ((c = *(tmp++)) != 0) {
	  if (verbose)
		  fprintf(stderr, "%s: %s: ", progname, errmsg);
	  switch(c) {
	  case 0x01:
		  if (verbose)
			  fprintf(stderr, "command not succeed.\n");
		  break;
	  case 0x02:
		  if (verbose)
			  fprintf(stderr, "CRC incorrect.\n");
		  break;
	  default:
		  if (verbose)
			  fprintf(stderr, "(code = 0x%02x)\n.",c);
	  }
	  return -2;
  }

  if(data == NULL)
    return 0;

  len--;
  for (i=0; i<len; i++) {
	  *data++ = *tmp++;
  }
  return len;
}

static int swra327_leds(PROGRAMMER *pgm, unsigned char command, unsigned char leds, unsigned char *data)
{
  unsigned char buffer[10];
  int result;

  buffer[0] = CMD_LEDS;
  buffer[1] = command;
  buffer[2] = leds;

  swra327_send(pgm, buffer, 3);

  result = swra327_vfy_cmd_sent(pgm, data, "leds command");
  if (result < 0) {
	  fprintf(stderr, "%s: error: leds command\n", progname);
	  return -1;
  }
  return 0;
}

static int swra327_err_led(PROGRAMMER *pgm, int command)
{
  return swra327_leds(pgm, command, LEDS_ERR, NULL);
}

static int swra327_vfy_led(PROGRAMMER *pgm, int command)
{
  return swra327_leds(pgm, command, LEDS_VFY, NULL);
}

/*
 * issue the 'chip erase' command to the device
 */
static int swra327_chip_erase(PROGRAMMER * pgm, PART * p)
{
  unsigned char c = CMD_ERASE;
  int result;

  swra327_send(pgm, &c, 1);
  result = swra327_vfy_cmd_sent(pgm, NULL, "chip erase");
  if (result < 0) {
	  fprintf(stderr, "%s: error: chip erase\n", progname);
	  return -1;
  }

  /*
   * swra327 firmware may not delay long enough
   */
  usleep (p->chip_erase_delay);

  return 0;
}

static void swra327_enter_prog_mode(PROGRAMMER * pgm)
{
  unsigned char c = CMD_HAND_SHAKE;
  int result;

  swra327_send(pgm, &c, 1);

  result = swra327_vfy_cmd_sent(pgm, NULL, "enter prog mode");
  if (result < 0) {
	  fprintf(stderr, "%s: error: enter prog mode\n", progname);
  }
}

/*
 * issue the 'program enable' command to the device
 */
static int swra327_program_enable(PROGRAMMER * pgm, PART * p)
{
  int result;
  unsigned char buf[10];

  buf[0] = CMD_ENABLE;
  buf[1] = buf[3] = 'O';
  buf[2] = buf[4] = 'K';
  swra327_send(pgm, buf, 5);

  result = swra327_vfy_cmd_sent(pgm, NULL, "prog enable");
  if (result < 0) {
	  fprintf(stderr, "%s: error: prog enable\n", progname);
	  return -1;
  }
  return 0;
}

static void swra327_leave_prog_mode(PROGRAMMER * pgm)
{
  unsigned char c = CMD_END;

  swra327_send(pgm, &c, 1);

  swra327_vfy_cmd_sent(pgm, NULL, "leave prog mode");
}


static int swra327_initialize(PROGRAMMER * pgm, PART * p)
{
  unsigned char c = SB_FORCE_BOOT;
  LNODEID ln;
  MEM *m;

  for (ln = lfirst(p->mem); ln; ln = lnext(ln)) {
    m = ldata(ln);
    if (strcmp(m->desc, "flash") == 0) {
    	if (m->page_size > 0)
    		PDATA(pgm)->flash_pagesize = m->page_size;
    }
  }
  free(PDATA(pgm)->flash_pagecache);
  if ((PDATA(pgm)->flash_pagecache = malloc(PDATA(pgm)->flash_pagesize)) == NULL) {
	  fprintf(stderr, "%s: swra327_initialize(): Out of memory\n",
			  progname);
	  return -1;
  }
  serial_send(&pgm->fd, &c, 1);

  swra327_enter_prog_mode(pgm);

  return 0;
}

static void swra327_disable(PROGRAMMER * pgm)
{
  /* Do nothing. */

  return;
}

static void swra327_enable(PROGRAMMER * pgm)
{
  /* Do nothing. */

  return;
}

/*
 * transmit a device command and return the results; 'cmd' and
 * 'res' must point to at least a 4 byte data buffer
 */
static int swra327_cmd(PROGRAMMER * pgm, unsigned char cmd[4],
		unsigned char res[4])
{
	return 0;
}

static int swra327_open(PROGRAMMER * pgm, char * port)
{
  /*
   *  If baudrate was not specified use 19.200 Baud
   */
  if(pgm->baudrate == 0) {
    pgm->baudrate = 19200;
  }

  strcpy(pgm->port, port);
  serial_open(port, pgm->baudrate, &pgm->fd);

  /*
   * drain any extraneous input
   */
  swra327_drain (pgm, 0);

  return 0;
}

static void swra327_close(PROGRAMMER * pgm)
{
  swra327_leave_prog_mode(pgm);

  serial_close(&pgm->fd);
  pgm->fd.ifd = -1;
}

static void swra327_display(PROGRAMMER * pgm, const char * p)
{
  return;
}
/*
static int _swra327_paged_load(PROGRAMMER * pgm, PART * p, int page_size, long n_bytes)
{
  int i, cmdlen, result;
  unsigned char *tmp, buf[275];
  unsigned char blocksize = 0;
  unsigned char *cache_ptr = NULL;
  int  n_block, n_tail;
  long saddr;

  MEM *mem;

  mem = locate_mem(p, "flash");
  saddr = part_mem_loaddr(mem);
  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_paged_load(..., %s, 0x%lx, 0x%lx, ...)\n",
			  progname, mem->desc, saddr, n_bytes);

  if (strcmp(mem->desc, "flash") != 0) {
 	  fprintf(stderr,
			  "%s: swra327_paged_load(): "
			  "unsupported memory type: %s\n",
			  progname, mem->desc);
	  return -1;
  }
  blocksize = mem->blocksize;
  if (blocksize == 0) {
	fprintf(stderr,
			"%s: swra327_paged_load(): "
			"zero block size error\n",
			progname);
	return -1;
  }
  n_block = (n_bytes - saddr) / blocksize;
  n_tail = (n_bytes -saddr) - (n_block * blocksize);

  for (i=0; i<n_block; i++) {
	  buf[0] = CMD_READ;
	  buf[1] = (saddr >> 16) & 0xff;
	  buf[2] = (saddr >> 8) & 0xff;
	  buf[3] = saddr & 0xff;
	  buf[4] = blocksize;
	  cmdlen = 5;

	  result = swra327_send(pgm, buf, cmdlen);

	  if (result < 0) {
		  fprintf(stderr,
				  "%s: swra327_paged_load(): timeout/error communicating with programmer\n",
				  progname);
		  return -1;
	  }

	  // attempt to get the data
	  result = swra327_vfy_cmd_sent(pgm, buf, "verify read command");

	  // check read back
	  if (result < 0) {
		  fprintf(stderr, "%s: swra327_paged_load(): error reading data\n",
				  progname);
		  return -1;
	  }
	  tmp = buf;
	  cache_ptr = mem->buf + saddr;
	  do {
		  *cache_ptr++ = *tmp++;
	  } while (blocksize--);
	  blocksize = mem->blocksize;
	  saddr += blocksize;
      report_progress(i, n_block, NULL);

  }
  if (n_tail) {
	  buf[0] = CMD_READ;
	  buf[1] = (saddr >> 16) & 0xff;
	  buf[2] = (saddr >> 8) & 0xff;
	  buf[3] = saddr & 0xff;
	  buf[4] = n_tail;
	  cmdlen = 5;

	  result = swra327_send(pgm, buf, cmdlen);

	  if (result < 0) {
		  fprintf(stderr,
				  "%s: swra327_paged_load(): timeout/error communicating with programmer\n",
				  progname);
		  return -1;
	  }

	  // attempt to get the data
	  result = swra327_vfy_cmd_sent(pgm, buf, "verify read command");

	  // check read back
	  if (result < 0) {
		  fprintf(stderr, "%s: swra327_paged_load(): error reading data\n",
				  progname);
		  return -1;
	  }
	  tmp = buf;
	  cache_ptr = mem->buf + saddr;
	  do {
		  *cache_ptr++ = *tmp++;
	  } while (n_tail--);
  }
  return 0;
}
*/

static int swra327_paged_load(PROGRAMMER * pgm, PART * p, int page_size, long n_bytes)
{
  int i, cmdlen, result;
  unsigned char *tmp, buf[275];
  unsigned char blocksize = 0;
  unsigned char *cache_ptr = NULL;
  int  n_block, n_tail;
  long saddr;
  int j;

  MEM *mem;

  mem = locate_mem(p, "flash");
  saddr = part_mem_loaddr(mem);
  if (verbose >= 2)
          fprintf(stderr, "%s: swra327_paged_load(..., %s, 0x%lx, 0x%lx, ...)\n",
                          progname, mem->desc, saddr, n_bytes);

  if (strcmp(mem->desc, "flash") != 0) {
          fprintf(stderr,
                          "%s: swra327_paged_load(): "
                          "unsupported memory type: %s\n",
                          progname, mem->desc);
          return -1;
  }
  blocksize = mem->blocksize;
  if (blocksize == 0) {
        fprintf(stderr,
                        "%s: swra327_paged_load(): "
                        "zero block size error\n",
                        progname);
        return -1;
  }
  n_block = (n_bytes - saddr) / blocksize;
  n_tail = (n_bytes -saddr) % blocksize;

  for (i=0; i<n_block; i++) {
    for (j=0; j<NUM_TRY; j++) {
      buf[0] = CMD_READ;
      buf[1] = (saddr >> 16) & 0xff;
      buf[2] = (saddr >> 8) & 0xff;
      buf[3] = saddr & 0xff;
      buf[4] = blocksize;
      cmdlen = 5;

      result = swra327_send(pgm, buf, cmdlen);

      if (result < 0) {
              fprintf(stderr,
                              "%s: swra327_paged_load(): timeout/error communicating with programmer\n",
                              progname);
              return -1;
      }

      // attempt to get the data
      result = swra327_vfy_cmd_sent(pgm, buf, "verify read command");

      // check read back
      if (result >= 0) {
        break;
      }
    }
    if (j == NUM_TRY)  {
            fprintf(stderr,
                            "\n\n%s: swra327_paged_load(): Unrecoverable error. \n",
                            progname);
            return -1;
    }
    tmp = buf;
          cache_ptr = mem->buf + saddr;
          do {
                  *cache_ptr++ = *tmp++;
          } while ((tmp-buf) < blocksize);
          saddr += blocksize;
          report_progress(i, n_block, NULL);

  }
  if (n_tail) {
    for (j=0; j<NUM_TRY; j++) {
      buf[0] = CMD_READ;
      buf[1] = (saddr >> 16) & 0xff;
      buf[2] = (saddr >> 8) & 0xff;
      buf[3] = saddr & 0xff;
      buf[4] = n_tail;
      cmdlen = 5;

      result = swra327_send(pgm, buf, cmdlen);

      if (result < 0) {
              fprintf(stderr,
                              "%s: swra327_paged_load(): timeout/error communicating with programmer\n",
                              progname);
              return -1;
      }

      // attempt to get the data
      result = swra327_vfy_cmd_sent(pgm, buf, "verify read command");

      // check read back
      if (result >= 0) {
        break;
      }
    }
    if (j == NUM_TRY)  {
            fprintf(stderr,
                            "\n\n%s: swra327_paged_load(): Unrecoverable error. \n",
                            progname);
            return -1;
    }
    tmp = buf;
    cache_ptr = mem->buf + saddr;
    do {
      *cache_ptr++ = *tmp++;
    } while ((tmp-buf) < n_tail);
  }
  return 0;
}

static int swra327_read_byte(PROGRAMMER * pgm, PART * p, unsigned long addr, unsigned char * value)
{
  int cmdlen, result;
  unsigned char buf[266];
  unsigned int blocksize = 0;

  MEM *mem;

  mem = locate_mem(p, "flash");

  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_read_byte(..., %s, 0x%lx, ...)\n",
			  progname, mem->desc, addr);

  if (strcmp(mem->desc, "flash") != 0) {
 	  fprintf(stderr,
			  "%s: swra327_read_byte(): "
			  "unsupported memory type: %s\n",
			  progname, mem->desc);
	  return -1;
  }

  buf[0] = CMD_READ;
  blocksize = mem->blocksize;
  if(blocksize == 0)
	  blocksize = 1;
  cmdlen = 5;
  buf[1] = (addr >> 16) & 0xff;
  buf[2] = (addr >> 8) & 0xff;
  buf[3] = addr & 0xff;
  buf[4] = 1;

  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_read_byte(): Sending read memory command: ",
			  progname);

  result = swra327_send(pgm, buf, cmdlen);

  if (result < 0) {
	  fprintf(stderr,
			  "%s: swra327_read_byte(): timeout/error communicating with programmer\n",
			  progname);
	  return -1;
  }

  // attempt to get the data
  result = swra327_vfy_cmd_sent(pgm, buf, "verify read command");

  // check read back
  if (result < 0) {
	  fprintf(stderr, "%s: swra327_read_byte(): error reading data\n",
			  progname);
	  return -1;
  }
  *value = *buf;
  return 0;
}

static int swra327_paged_write(PROGRAMMER * pgm, PART * p, int page_size, long n_bytes)
{
  int i, j, cmdlen, result;
  unsigned char c, buf[275];
  unsigned int blocksize = 0;
  unsigned char *cache_ptr = NULL;
  int  n_block, n_tail;
  long saddr;

  MEM *mem;

  mem = locate_mem(p, "flash");
  saddr = part_mem_loaddr(mem);
  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_paged_write(.., %s, 0x%lx, 0x%lx, ...)\n",
			  progname, mem->desc, saddr, n_bytes);
  if (strcmp(mem->desc, "flash") != 0) {
 	  fprintf(stderr,
			  "%s: swra327_paged_write(): "
			  "unsupported memory type: %s\n",
			  progname, mem->desc);
	  return -1;
  }
  blocksize = mem->blocksize;
  if (blocksize == 0) {
	fprintf(stderr,
			"%s: swra327_paged_write(): "
			"zero block size error\n",
			progname);
	return -1;
  }
  cache_ptr = mem->buf + saddr;
  n_block = (n_bytes - saddr) / blocksize;
  n_tail = (n_bytes -saddr) % blocksize;

  for (i=0; i<n_block; i++) {
	for(j=0; j<NUM_TRY; j++) {
		buf[0] = CMD_WRITE;
		buf[1] = (saddr >> 16) & 0xff;
		buf[2] = (saddr >> 8) & 0xff;
		buf[3] = saddr & 0xff;
		buf[4] = blocksize;
		cmdlen = blocksize + 5;
		memcpy(buf + 5, cache_ptr, blocksize);
		result = swra327_send(pgm, buf, cmdlen);
		if (result < 0) {
			fprintf(stderr,
					"%s: swra327_paged_write(): "
					"timeout/error communicating with programmer\n",
					progname);
			return -1;
		}

		result = swra327_vfy_cmd_sent(pgm, &c, "swra327_paged_write");
		// check read back
		if (result >= 0) {
			break;
		}
	}
	if (j == NUM_TRY)  {
		fprintf(stderr,
				"\n\n%s: swra327_paged_write(): Unrecoverable error. \n",
				progname);
		return -1;
	}

	cache_ptr += blocksize;
	saddr += blocksize;
	report_progress(i, n_block, NULL);
  }
  if (n_tail) {
	if(n_tail % 4) {
		n_tail += 4-(n_tail % 4);
	}
	for (j=0; j<NUM_TRY; j++) {
		buf[0] = CMD_WRITE;
		buf[1] = (saddr >> 16) & 0xff;
		buf[2] = (saddr >> 8) & 0xff;
		buf[3] = saddr & 0xff;
		buf[4] = n_tail;
		cmdlen = n_tail + 5;
		memcpy(buf + 5, cache_ptr, n_tail);

		result = swra327_send(pgm, buf, cmdlen);
		if (result < 0) {
			fprintf(stderr,
					"%s: swra327_paged_write(): "
					"timeout/error communicating with programmer\n",
					progname);
			return -1;
		}

		result = swra327_vfy_cmd_sent(pgm, &c, "verify write command");
		// check read back
		if (result >= 0) {
			break;
		}
	}
	if (j == NUM_TRY)  {
		fprintf(stderr,
				"\n\n%s: swra327_paged_write(): Unrecoverable error. \n",
				progname);
		return -1;
	}
  }
  return n_bytes;
}

static int swra327_write_byte(PROGRAMMER * pgm, PART * p, unsigned long addr, unsigned char data)
{
  int j, cmdlen, result;
  unsigned char c, buf[275];
  unsigned int blocksize = 0;
  unsigned char *cache_ptr = NULL;
  MEM *mem;

  mem = p->mem;
  mem = locate_mem(p, "flash");
  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_write_byte(.., %s, 0x%lx, ...)\n",
			  progname, mem->desc, addr);
  if (strcmp(mem->desc, "flash") != 0) {
 	  fprintf(stderr,
			  "%s: swra327_write_byte(): "
			  "unsupported memory type: %s\n",
			  progname, mem->desc);
	  return -1;
  }
  memset(buf, 0xff, sizeof(buf));
  buf[0] = CMD_WRITE;
  blocksize = mem->blocksize;
  cache_ptr = PDATA(pgm)->flash_pagecache;

  if (blocksize) {
	  cmdlen = blocksize + 3;
	  memset(cache_ptr, 0xff, blocksize);
	  cache_ptr[addr & (blocksize - 1)] = data;
	  addr &= ~(blocksize-1);
	  buf[1] = (addr >> 16) & 0xff;
	  buf[2] = (addr >> 8) & 0xff;
	  buf[3] = addr & 0xff;
	  buf[4] = blocksize;
	  memcpy(buf + 5, cache_ptr, blocksize);
  } else {
	  cmdlen = 5 + 3;
	  buf[1] = (addr >> 16) & 0xff;
	  buf[2] = (addr >> 8) & 0xff;
	  buf[3] = addr & 0xff;
	  buf[4] = 4;    // length must be multiple of 4
	  buf[5] = data;
  }
  if (verbose >= 2)
	  fprintf(stderr, "%s: swra327_write_byte(): Sending write memory command: ",
			  progname);

  for (j=0; j<NUM_TRY; j++) {
	  result = swra327_send(pgm, buf, cmdlen);
	  if (result < 0) {
		  fprintf(stderr,
				  "%s: swra327_write_byte(): "
				  "timeout/error communicating with programmer\n",
				  progname);
		  return -1;
	  }
	  result = swra327_vfy_cmd_sent(pgm, &c, "swra327_write_byte");
	  // check read back
	  if (result >= 0) {
		  break;
	  }
	}
	if (j == NUM_TRY)  {
		fprintf(stderr,
				"\n\n%s: swra327_write_byte(): Unrecoverable error. \n",
				progname);
		return -1;
	}
	return 0;
}

static int swra327_read_sig_bytes(PROGRAMMER * pgm, PART * p)
{
  int cmdlen, result;
  unsigned char buf[255];
  MEM *sig;

  buf[0] = CMD_READ_SIG;
  cmdlen = 1;

  sig = locate_mem(p, "signature");
  if (sig == NULL) {
    fprintf(stderr,
            "%s: WARNING: signature data not defined for device \"%s\"\n",
            progname, p->desc);
    return -1;
  }


  result = swra327_send(pgm, buf, cmdlen);
  if (result < 0) {
	  fprintf(stderr,
			  "%s: swra327_read_sig_bytes(): "
			  "timeout/error communicating with programmer\n",
			  progname);
	  return -1;
  }
  return swra327_vfy_cmd_sent(pgm, sig->buf, "swra327_read_sig_bytes");
}

void swra327_initpgm(PROGRAMMER * pgm)
{
  strcpy(pgm->type, "SWRA327");

  /*
   * mandatory functions
   */
  pgm->initialize     = swra327_initialize;
  pgm->display        = swra327_display;
  pgm->enable         = swra327_enable;
  pgm->disable        = swra327_disable;
  pgm->program_enable = swra327_program_enable;
  pgm->chip_erase     = swra327_chip_erase;
  pgm->cmd            = swra327_cmd;
  pgm->open           = swra327_open;
  pgm->close          = swra327_close;

  /*
   * optional functions
   */
  pgm->write_byte     = swra327_write_byte;
  pgm->read_byte      = swra327_read_byte;
  pgm->paged_write    = swra327_paged_write;
  pgm->paged_load     = swra327_paged_load;
  pgm->setup          = swra327_setup;
  pgm->teardown       = swra327_teardown;
  pgm->read_sig_bytes = swra327_read_sig_bytes;
  pgm->err_led        = swra327_err_led;
  pgm->vfy_led        = swra327_vfy_led;
}
